import 'rxjs/add/operator/toPromise';
import { AsyncSubject as AsyncSubject$1 } from 'rxjs/AsyncSubject';
import { Observable as Observable$1 } from 'rxjs/Observable';
import { Inject, InjectionToken, Injector, NgModule, Optional, Pipe } from '@angular/core';
import { BehaviorSubject as BehaviorSubject$1 } from 'rxjs/BehaviorSubject';
import { ReplaySubject as ReplaySubject$1 } from 'rxjs/ReplaySubject';
import { Subject as Subject$1 } from 'rxjs/Subject';
import { CommonModule } from '@angular/common';
import { createStore } from 'redux';

class MetadataManager {
    /**
     * @template T
     * @param {?} target
     * @param {?} propertyKey
     * @return {?}
     */
    static getPropertyType(target, propertyKey) {
        const /** @type {?} */ metadata = Reflect['getMetadata']('design:type', target, propertyKey);
        const /** @type {?} */ fallback = '';
        if (metadata) {
            return metadata.name || fallback;
        }
        return fallback;
    }
    /**
     * @template T
     * @param {?} target
     * @param {?} data
     * @return {?}
     */
    static setActionMetadata(target, data) {
        MetadataManager.set(target, MetadataManager.ACTION_KEY, data);
    }
    /**
     * @template T
     * @param {?} target
     * @return {?}
     */
    static getActionMetadata(target) {
        return MetadataManager.get(target, MetadataManager.ACTION_KEY, MetadataManager.ACTION_DEFAULT);
    }
    /**
     * @template T
     * @param {?} target
     * @param {?} data
     * @return {?}
     */
    static setActionContextMetadata(target, data) {
        MetadataManager.set(target, MetadataManager.ACTION_CONTEXT_KEY, data);
    }
    /**
     * @template T
     * @param {?} target
     * @return {?}
     */
    static getActionContextMetadata(target) {
        return MetadataManager.get(target, MetadataManager.ACTION_CONTEXT_KEY, MetadataManager.ACTION_CONTEXT_DEFAULT);
    }
    /**
     * @template T
     * @param {?} target
     * @param {?} data
     * @return {?}
     */
    static setReducerMetadata(target, data) {
        MetadataManager.set(target, MetadataManager.REDUCER_KEY, data);
    }
    /**
     * @template T
     * @param {?} target
     * @return {?}
     */
    static getReducerMetadata(target) {
        return MetadataManager.get(target, MetadataManager.REDUCER_KEY, MetadataManager.REDUCER_DEFAULT);
    }
    /**
     * @template T
     * @param {?} target
     * @param {?} data
     * @return {?}
     */
    static setStateMetadata(target, data) {
        MetadataManager.set(target, MetadataManager.STATE_KEY, data);
    }
    /**
     * @template T
     * @param {?} target
     * @return {?}
     */
    static getStateMetadata(target) {
        return MetadataManager.get(target, MetadataManager.STATE_KEY, MetadataManager.STATE_DEFAULT);
    }
    /**
     * @template D, T
     * @param {?} target
     * @param {?} key
     * @param {?} data
     * @return {?}
     */
    static set(target, key, data) {
        Reflect['defineMetadata'](key, data, target);
    }
    /**
     * @template D, T
     * @param {?} target
     * @param {?} key
     * @param {?} defaultData
     * @return {?}
     */
    static get(target, key, defaultData) {
        defaultData = defaultData || /** @type {?} */ ({});
        return Object.assign({}, defaultData, Reflect['getMetadata'](key, target));
    }
}
MetadataManager.ACTION_KEY = Symbol('@ReduxAction');
MetadataManager.ACTION_DEFAULT = {
    contextClazz: null,
    type: '',
};
MetadataManager.ACTION_CONTEXT_KEY = Symbol('@ReduxActionContext');
MetadataManager.ACTION_CONTEXT_DEFAULT = {
    prefix: '',
};
MetadataManager.REDUCER_KEY = Symbol('@ReduxReducer');
MetadataManager.REDUCER_DEFAULT = {
    reducers: [],
};
MetadataManager.STATE_KEY = Symbol('@ReduxState');
MetadataManager.STATE_DEFAULT = {
    name: '',
};

/**
 * @param {?} config
 * @return {?}
 */
function ReduxActionContextDecorator(config) {
    return (constructor) => {
        MetadataManager.setActionContextMetadata(constructor, config);
        return constructor;
    };
}

class ReduxRegistry {
    /**
     * @param {?} store
     * @return {?}
     */
    static registerStore(store) {
        ReduxRegistry._store.next(store);
        ReduxRegistry._store.complete();
    }
    /**
     * @param {?} stateName
     * @param {?} types
     * @param {?} reducer
     * @return {?}
     */
    static registerReducer(stateName, types, reducer) {
        if (Array.isArray(types)) {
            types.forEach((type) => {
                ReduxRegistry._reducers.push({
                    reducer,
                    stateName,
                    type,
                });
            });
        }
        else {
            ReduxRegistry._reducers.push({
                reducer,
                stateName,
                type: /** @type {?} */ (types),
            });
        }
    }
    /**
     * @param {?} state
     * @return {?}
     */
    static registerState(state) {
        ReduxRegistry.getStore().then((store) => {
            const /** @type {?} */ stateConfig = MetadataManager.getStateMetadata(state.constructor);
            const /** @type {?} */ initState = state.getInitialState();
            let /** @type {?} */ initStateToResolve = initState;
            if (initState instanceof Observable$1) {
                initStateToResolve = initState.toPromise();
            }
            Promise.resolve(initStateToResolve).then((initialState) => {
                store.dispatch({
                    payload: {
                        initialState,
                        stateName: stateConfig.name,
                    },
                    type: ReduxRegistry.ACTION_REGISTER_MODULE,
                });
            });
        });
    }
    /**
     * @param {?} type
     * @return {?}
     */
    static getReducerItemsByType(type) {
        return ReduxRegistry._reducers
            .filter(reducerItem => getActionTypeByFunction(reducerItem.type) === type);
    }
    /**
     * @return {?}
     */
    static getStore() {
        return new Promise(ReduxRegistry._store.subscribe.bind(ReduxRegistry._store));
    }
}
ReduxRegistry.ACTION_REGISTER_MODULE = `${ReduxRegistry.name}://registerModule`;
ReduxRegistry._store = new AsyncSubject$1();
ReduxRegistry._reducers = [];

/**
 * @param {?=} config
 * @return {?}
 */
function ReduxActionDecorator(config) {
    return (target, propertyKey, descriptor) => {
        config = config || { type: propertyKey };
        const /** @type {?} */ originalFunction = descriptor.value;
        const /** @type {?} */ proxyFunction = function () {
            const /** @type {?} */ contextData = MetadataManager.getActionContextMetadata(target.constructor);
            const /** @type {?} */ type = getActionType(contextData.prefix, config.type);
            const /** @type {?} */ returnValue = originalFunction.apply(this, arguments);
            Promise.all([
                Promise.resolve(returnValue),
                ReduxRegistry.getStore(),
            ]).then(([payload, store]) => {
                store.dispatch({ payload, type });
            });
            return returnValue;
        };
        MetadataManager.setActionMetadata(proxyFunction, {
            contextClazz: target.constructor,
            type: config.type,
        });
        descriptor.value = proxyFunction;
    };
}
/**
 * @param {?} target
 * @return {?}
 */
function getActionTypeByFunction(target) {
    const { type, contextClazz } = MetadataManager.getActionMetadata(target);
    const { prefix } = MetadataManager.getActionContextMetadata(contextClazz);
    return getActionType(prefix, type);
}
/**
 * @param {...?} rest
 * @return {?}
 */
function getActionType(...rest) {
    return rest.filter((i) => i.length > 0).join('/');
}

class ReduxModuleRootReducer {
    /**
     * @param {?} state
     * @param {?} action
     * @return {?}
     */
    static reduce(state, action) {
        if (action.type === ReduxRegistry.ACTION_REGISTER_MODULE) {
            return ReduxModuleRootReducer.reduceRegisterModule(state, /** @type {?} */ (action));
        }
        return ReduxModuleRootReducer.reduceByRegisteredReducers(state, action);
    }
    /**
     * @param {?} state
     * @param {?} action
     * @return {?}
     */
    static reduceRegisterModule(state, action) {
        const { stateName, initialState } = action.payload;
        return Object.assign({}, state, {
            [stateName]: initialState,
        });
    }
    /**
     * @param {?} state
     * @param {?} action
     * @return {?}
     */
    static reduceByRegisteredReducers(state, action) {
        const /** @type {?} */ reducers = ReduxRegistry.getReducerItemsByType(action.type);
        return reducers.reduce((stateToReduce, reducerItem) => {
            const { stateName, reducer } = reducerItem;
            return Object.assign({}, stateToReduce, {
                [stateName]: reducer(stateToReduce[stateName], action),
            });
        }, state);
    }
}

/**
 * @template S, P
 * @param {?} types
 * @return {?}
 */
function ReduxReducerDecorator(types) {
    return (target, propertyKey) => {
        const /** @type {?} */ data = MetadataManager.getReducerMetadata(target.constructor);
        data.reducers.push({ types, reducer: target[propertyKey].bind(target) });
        MetadataManager.setReducerMetadata(target.constructor, data);
        return target[propertyKey];
    };
}

const REDUX_MODULE_CHILD_CONFIG = new InjectionToken('REDUX_MODULE_CHILD_CONFIG');

let ReduxStateSelectorDataType = {};
ReduxStateSelectorDataType.SUBJECT = 0;
ReduxStateSelectorDataType.BEHAVIOR_SUBJECT = 1;
ReduxStateSelectorDataType.REPLAY_SUBJECT = 2;
ReduxStateSelectorDataType.OBSERVABLE = 3;
ReduxStateSelectorDataType[ReduxStateSelectorDataType.SUBJECT] = "SUBJECT";
ReduxStateSelectorDataType[ReduxStateSelectorDataType.BEHAVIOR_SUBJECT] = "BEHAVIOR_SUBJECT";
ReduxStateSelectorDataType[ReduxStateSelectorDataType.REPLAY_SUBJECT] = "REPLAY_SUBJECT";
ReduxStateSelectorDataType[ReduxStateSelectorDataType.OBSERVABLE] = "OBSERVABLE";

class ReduxStateSelector {
    /**
     * @param {?} expression
     * @param {?=} context
     */
    constructor(expression, context) {
        this.expression = expression;
        this.context = context;
        if (!expression.startsWith(ReduxStateSelector.DELIMITER)) {
            if (!context) {
                throw new Error('You need to provide a state context if you want to use relativ selectors');
            }
            this.expression = `/${MetadataManager.getStateMetadata(context).name}/${this.expression}`;
        }
    }
    /**
     * @template S
     * @return {?}
     */
    asObservable() {
        return /** @type {?} */ (this.getByDataType(ReduxStateSelectorDataType.OBSERVABLE, null));
    }
    /**
     * @template S
     * @return {?}
     */
    asBehaviorSubject() {
        return /** @type {?} */ (this.getByDataType(ReduxStateSelectorDataType.BEHAVIOR_SUBJECT, null));
    }
    /**
     * @template S
     * @return {?}
     */
    asReplaySubject() {
        return /** @type {?} */ (this.getByDataType(ReduxStateSelectorDataType.REPLAY_SUBJECT, null));
    }
    /**
     * @template S
     * @return {?}
     */
    asSubject() {
        return /** @type {?} */ (this.getByDataType(ReduxStateSelectorDataType.SUBJECT, null));
    }
    /**
     * @template S
     * @param {?} dataType
     * @param {?} initialValue
     * @return {?}
     */
    getByDataType(dataType, initialValue) {
        const /** @type {?} */ subject = this.getSubjectByDataType(dataType, initialValue);
        ReduxRegistry.getStore().then(store => {
            subject.next(this.getValueFromState(store.getState()));
            store.subscribe(() => subject.next(this.getValueFromState(store.getState())));
        });
        if (dataType === ReduxStateSelectorDataType.OBSERVABLE) {
            return subject.asObservable();
        }
        return subject;
    }
    /**
     * @template T, S
     * @param {?} state
     * @return {?}
     */
    getValueFromState(state) {
        return this.expression.split(ReduxStateSelector.DELIMITER)
            .filter(ReduxStateSelector.isPropertyKeyValid)
            .reduce((previousValue, propertyKey) => {
            if (!previousValue || !previousValue[propertyKey]) {
                return null;
            }
            return previousValue[propertyKey];
        }, state);
    }
    /**
     * @template S
     * @param {?} dataType
     * @param {?} initialValue
     * @return {?}
     */
    getSubjectByDataType(dataType, initialValue) {
        switch (dataType) {
            case ReduxStateSelectorDataType.REPLAY_SUBJECT:
                return new ReplaySubject$1();
            case ReduxStateSelectorDataType.SUBJECT:
                return new Subject$1();
            default:
                return new BehaviorSubject$1(initialValue);
        }
    }
    /**
     * @param {?} propertyKey
     * @return {?}
     */
    static isPropertyKeyValid(propertyKey) {
        return propertyKey !== '';
    }
}
ReduxStateSelector.DELIMITER = '/';

class ReduxSelectPipe {
    /**
     * @param {?} moduleConfig
     */
    constructor(moduleConfig) {
        this.moduleConfig = moduleConfig;
    }
    /**
     * @param {?} selector
     * @return {?}
     */
    transform(selector) {
        return new ReduxStateSelector(selector, this.moduleConfig.state).asObservable();
    }
}
ReduxSelectPipe.decorators = [
    { type: Pipe, args: [{ name: 'reduxSelect' },] },
];
/**
 * @nocollapse
 */
ReduxSelectPipe.ctorParameters = () => [
    { type: undefined, decorators: [{ type: Inject, args: [REDUX_MODULE_CHILD_CONFIG,] },] },
];

/**
 * @param {?} dataType
 * @return {?}
 */
function reduxStateSelectorDataTypeFactory(dataType) {
    return {
        BehaviorSubject: ReduxStateSelectorDataType.BEHAVIOR_SUBJECT,
        Observable: ReduxStateSelectorDataType.OBSERVABLE,
        ReplaySubject: ReduxStateSelectorDataType.REPLAY_SUBJECT,
        Subject: ReduxStateSelectorDataType.SUBJECT,
    }[dataType];
}

/**
 * @param {?} expression
 * @param {?=} context
 * @return {?}
 */
function ReduxSelectDecorator(expression, context) {
    return (target, propertyKey) => {
        const /** @type {?} */ propertyType = MetadataManager.getPropertyType(target, propertyKey);
        Object.defineProperty(target, propertyKey, {
            configurable: true,
            enumerable: true,
            value: new ReduxStateSelector(expression, context)
                .getByDataType(reduxStateSelectorDataTypeFactory(propertyType), null),
        });
    };
}

/**
 * @param {?} config
 * @return {?}
 */
function ReduxStateDecorator(config) {
    return (constructor) => {
        MetadataManager.setStateMetadata(constructor, config);
        return constructor;
    };
}

const REDUX_MODULE_ROOT_CONFIG = new InjectionToken('REDUX_MODULE_ROOT_CONFIG');

const IS_ROOT_MODULE = new InjectionToken('IS_ROOT_MODULE');

class ReduxModule {
    /**
     * @param {?=} isRootModule
     * @param {?=} rootConfig
     * @param {?=} childConfig
     * @param {?=} injector
     */
    constructor(isRootModule = false, rootConfig = null, childConfig = null, injector) {
        this.injector = injector;
        if (isRootModule) {
            rootConfig = Object.assign({
                store: null,
            }, rootConfig || {});
            ReduxRegistry.registerStore(rootConfig.store || createStore(ReduxModuleRootReducer.reduce, {}));
        }
        if (childConfig) {
            this.initChildConfig(childConfig);
        }
    }
    /**
     * @param {?} childConfig
     * @return {?}
     */
    initChildConfig(childConfig) {
        if (childConfig.state) {
            ReduxRegistry.registerState(this.injector.get(childConfig.state));
            if (childConfig.reducers) {
                const /** @type {?} */ stateMetadata = MetadataManager.getStateMetadata(childConfig.state);
                childConfig.reducers
                    .map((reducer) => MetadataManager.getReducerMetadata(reducer.constructor))
                    .forEach((metadata) => {
                    metadata.reducers.forEach(reducer => {
                        ReduxRegistry.registerReducer(stateMetadata.name, reducer.types, reducer.reducer);
                    });
                });
            }
        }
    }
    /**
     * @param {?} config
     * @return {?}
     */
    static forChild(config) {
        return {
            ngModule: ReduxModule,
            providers: [
                config.state,
                { provide: IS_ROOT_MODULE, useValue: false },
                { provide: REDUX_MODULE_CHILD_CONFIG, useValue: config },
            ],
        };
    }
    /**
     * @param {?=} config
     * @return {?}
     */
    static forRoot(config) {
        return {
            ngModule: ReduxModule,
            providers: [
                { provide: IS_ROOT_MODULE, useValue: true },
                { provide: REDUX_MODULE_ROOT_CONFIG, useValue: config },
            ],
        };
    }
}
ReduxModule.decorators = [
    { type: NgModule, args: [{
                declarations: [
                    ReduxSelectPipe,
                ],
                exports: [
                    ReduxSelectPipe,
                ],
                imports: [
                    CommonModule,
                ],
            },] },
];
/**
 * @nocollapse
 */
ReduxModule.ctorParameters = () => [
    { type: undefined, decorators: [{ type: Inject, args: [IS_ROOT_MODULE,] },] },
    { type: undefined, decorators: [{ type: Optional }, { type: Inject, args: [REDUX_MODULE_ROOT_CONFIG,] },] },
    { type: undefined, decorators: [{ type: Optional }, { type: Inject, args: [REDUX_MODULE_CHILD_CONFIG,] },] },
    { type: Injector, },
];

/**
 * Generated bundle index. Do not edit.
 */

export { ReduxActionContextDecorator as ReduxActionController, ReduxActionDecorator as ReduxAction, ReduxModuleRootReducer, ReduxReducerDecorator as ReduxReducer, ReduxSelectDecorator as ReduxSelect, ReduxSelectPipe, ReduxStateDecorator as ReduxState, ReduxStateSelector, ReduxModule, REDUX_MODULE_CHILD_CONFIG as ɵa, REDUX_MODULE_ROOT_CONFIG as ɵd, IS_ROOT_MODULE as ɵc };
//# sourceMappingURL=ngx-redux-core.js.map
